<?php

/**
 * Authentication Plugin: OpenID Single Sign-On (SSO) Authentication
 *
 * This plugin provides OpenID SSO functionality in Moodle by making use of the
 * Identifier Select feature in OpenID 2.0 to authenticate against fixed (and
 * ideally trusted) provider.  Once enabled, you must configure a fixed server
 * in the OpenID SSO auth settings.  This should only be done when at least one
 * OpenID user has been assigned full administrative rights to avoid being
 * locked out of the site.
 *
 * If you do happen to get locked out of your site, uncomment the line in the
 * class constructor which looks like this:
 *
 * <code>
 * // If you need to disable this plugin manually, uncomment the next line
 * // $this->config->openid_sso_url='';
 * </code>
 *
 * @author Stuart Metcalfe <info@pdl.uk.com>
 * @copyright Copyright (c) 2007 Canonical
 * @license http://www.gnu.org/copyleft/gpl.html GNU Public License
 * @package moodle multiauth
 */

if (!defined('MOODLE_INTERNAL')) {
    die('Direct access to this script is forbidden.');    ///  It must be included from a Moodle page
}

require_once $CFG->libdir.'/authlib.php';
require_once $CFG->dirroot.'/mod/openid/lib.php';

// Append the OpenID directory to the include path and include relevant files
set_include_path(get_include_path().PATH_SEPARATOR.$CFG->libdir.'/openid/');

// Required files (library)
require_once 'Auth/OpenID/Consumer.php';
require_once 'Auth/OpenID/FileStore.php';

// Include the custom event script if it exists
if (file_exists($CFG->dirroot.'/mod/openid/event.php')) {
    include $CFG->dirroot.'/mod/openid/event.php';
}

/**
 * OpenID SSO authentication plugin.
 */
class auth_plugin_openid_sso extends auth_plugin_base {

    /**
     * Class constructor
     *
     * Assigns default config values and checks for requested actions
     */
    function auth_plugin_openid_sso() {
        global $USER;
        
        $this->authtype = 'openid_sso';
        $this->config = get_config('auth/openid');
        
        // Set some defaults if not already set up
        if (!isset($this->config->openid_sso_url)) {
            set_config('openid_sso_url', '', 'auth/openid');
            $this->config->openid_sso_url='';
        }
        
        // If you need to disable this plugin manualy, uncomment the next line
        // $this->config->openid_sso_url='';
        
        if (!isset($this->config->openid_sreg_required)) {
            set_config('openid_sreg_required', 'nickname,email,fullname', 'auth/openid');
            $this->config->openid_sreg_required='nickname,email,fullname';
        }
        
        if (!isset($this->config->openid_sreg_optional)) {
            set_config('openid_sreg_optional', 'country', 'auth/openid');
            $this->config->openid_sreg_optional='country';
        }
        
        if (!isset($this->config->openid_privacy_url)) {
            set_config('openid_privacy_url', '', 'auth/openid');
            $this->config->openid_privacy_url='';
        }
        
        // Define constants used in OpenID lib
        define('OPENID_USE_IDENTIFIER_SELECT', 'true');
    }
    
    /**
     * This is the primary method that is used by the authenticate_user_login()
     * function in moodlelib.php. This method should return a boolean indicating
     * whether or not the username and password authenticate successfully.
     *
     * Returns true if the username and password work and false if they are
     * wrong or don't exist.
     *
     * @param string $username The username (with system magic quotes)
     * @param string $password The password (with system magic quotes)
     *
     * @return bool Authentication success or failure.
     */
    function user_login($username, $password) {
        // This plugin doesn't use usernames and passwords
        return false;
    }

    /**
     * Prints a form for configuring this authentication plugin.
     *
     * This function is called from admin/auth.php, and outputs a full page with
     * a form for configuring this plugin.
     */
    function config_form($config, $err, $user_fields) {
        global $CFG, $USER;
        include $CFG->dirroot.'/mod/openid/auth_config_sso.html';
    }

    /**
     * A chance to validate form data, and last chance to
     * do stuff before it is inserted in config_plugin
     * @param object object with submitted configuration settings (without system magic quotes)
     * @param array $err array of error messages
     */
    function validate_form(&$form, &$err) {
        //override if needed
    }

    /**
     * Processes and stores configuration data for this authentication plugin.
     *
     * @param object object with submitted configuration settings (without system magic quotes)
     */
    function process_config($config) {
        // save settings
        $vars = array(
            'openid_sso_url',
            'openid_sreg_required',
            'openid_sreg_optional',
            'openid_privacy_url'
        );
        
        foreach ($vars as $var) {
            set_config($var, $config->$var, 'auth/openid');
        }
        
        return true;
    }

    /**
     * Hook for overriding behavior of login page.
     * This method is called from login/index.php page for all enabled auth
     * plugins.
     *
     * We're overriding the default login behaviour.
     */
    function loginpage_hook() {
        global $CFG;
        global $frm, $user; // Login page variables
        
        // If the OpenID server isn't defined, don't do anything
        if (empty($this->config->openid_sso_url)) {
            return;
        }
        
        $mode = optional_param('openid_mode', null);
        $allow_append = ($this->config->auth_openid_allow_muliple=='true');
        
        if ($mode == null) {
            // If openid.mode isn't defined, initiate a request
            $this->do_request();
        } else {
            // Otherwise, we'll assume this is a response
            $resp = $this->process_response($_GET);
            
            if ($resp !== false) {
                $url = $resp->identity_url;
                
                if (record_exists('openid_urls', 'url', $url)) {
                    // Get the user associated with the OpenID
                    $userid = get_field('openid_urls', 'userid', 'url', $url);
                    $user = get_complete_user_data('id', $userid);
                    
                    // If the user isn't found then there's a database
                    // discrepancy.  We delete this entry and create a new user
                    if (!$user) {
                        delete_records('openid_urls', 'url', $url);
                        $user = $this->_create_account($resp);
                    }
                    
                    // Otherwise, the user is found and we call the optional
                    // on_openid_login function
                    elseif (function_exists('on_openid_login')) {
                        on_openid_login($resp, $user);
                    }
                } else {
                    // Otherwise, create a new account
                    $user = $this->_create_account($resp);
                }
                
                $frm->username = $user->username;
                $frm->password = $user->password;
            }
        }
    }
    
    /**
     * Initiate an OpenID request
     *
     * @param boolean $allow_sreg Default true
     * @param string $process_url Default empty (will use $CFG->wwwroot)
     * @param array $params Array of extra parameters to append to the request
     */
    function do_request($allow_sreg=true, $process_url='', $params=array()) {
        global $CFG;
        
        // Create the consumer instance
        $store = new Auth_OpenID_FileStore($CFG->dirroot.'/mod/openid/store');
        $consumer = new Auth_OpenID_Consumer($store);
        
        // Create our own endpoint and skip the discovery step.
        $endpoint = new Auth_OpenID_ServiceEndpoint();
        $endpoint->server_url = $this->config->openid_sso_url;
        $endpoint->identity_url = OPENID_IDENTIFIER_SELECT_NS;
        $endpoint->type_uris = array(_OPENID_1_0_TYPE);
        $authreq = $consumer->beginWithoutDiscovery($endpoint);
        
        if (!$authreq) {
            error(get_string('auth_openid_login_error', 'openid'));
        } else {
            // Add any simple registration fields to the request
            if ($allow_sreg === true) {
                $sreg_added = false;
                
                // Required fields
                if (!empty($this->config->openid_sreg_required)) {
                    $req = $this->config->openid_sreg_required;
                    $authreq->addExtensionArg('sreg', 'required', $req);
                    $sreg_added = true;
                }
                
                // Optional fields
                if (!empty($this->config->openid_sreg_optional)) {
                    $opt = $this->config->openid_sreg_optional;
                    $authreq->addExtensionArg('sreg', 'optional', $opt);
                    $sreg_added = true;
                }
                
                // Privacy statement
                if ($sreg_added && !empty($this->config->openid_privacy_url)) {
                    $url = $this->config->openid_privacy_url;
                    $authreq->addExtensionArg('sreg', 'policy_url', $url);
                }
            }
            
            // Finally, redirect to the OpenID provider
            if (empty($process_url)) {
                $process_url = $CFG->wwwroot.'/login/index.php';
            }
            
            if (is_array($params) && !empty($params)) {
                $query = '';
                
                foreach ($params as $key=>$val) {
                    $query .= '&'.$key.'='.$val;
                }
                
                $process_url .= '?'.substr($query, 1);
            }
            
            $trust_root = $CFG->wwwroot.'/';
            $redirect_url = $authreq->redirectURL($trust_root, $process_url);
            redirect($redirect_url);
        }
    }
    
    /**
     * Process an OpenID response
     *
     * By default, this method uses the error() function to display errors. This
     * is a terminal function so if you want to display errors inline using the
     * notify() function you will need to pass true to the $notify_errors
     * argument
     *
     * @param array $resp_data Most likely $_GET
     * @param boolean $notify_errors Default true
     * @return mixed Successful response object or false
     */
    function process_response($resp_data, $notify_errors=false) {
        global $CFG;
        
        // Create the consumer instance
        $store = new Auth_OpenID_FileStore($CFG->dirroot.'/mod/openid/store');
        $consumer = new Auth_OpenID_Consumer($store);
        $resp = $consumer->complete($resp_data);
        
        // Act based on response status
        switch ($resp->status) {
        case Auth_OpenID_SUCCESS:
            // Auth succeeded
            return $resp;
        
        case Auth_OpenID_CANCEL:
            // Auth cancelled by user.
            $msg = get_string('auth_openid_user_cancelled', 'openid');
            
            if ($notify_errors) {
                notify($msg);
            } else {
                error($msg);
            }
            
            break;
        
        case Auth_OpenID_FAILURE:
            // Auth failed for some reason
            $msg = openid_get_friendly_message($resp->message);
            $msg = get_string('auth_openid_login_failed', 'openid', $msg);
            
            if ($notify_errors) {
                notify($msg);
            } else {
                error($msg);
            }
        }
        
        return false;
    }
    
    /**
     * Create a new account using simple registration data if available
     *
     * @access private
     * @param object &$resp An OpenID consumer response object
     * @return object The new user
     */
    function _create_account(&$resp) {
        global $CFG, $USER;
        
        $url = $resp->identity_url;
        $password = hash_internal_user_password('openid');
        $server = $resp->endpoint->server_url;
        $sreg = $resp->extensionResponse('sreg');
        
        // We'll attempt to use the user's nickname to set their username
        if (isset($sreg['nickname']) && !empty($sreg['nickname']) && !record_exists('users', 'username', $sreg['nickname'])) {
            $username = $sreg['nickname'];
        }
        
        // Otherwise, we'll use their openid url
        else {
            $username = openid_normalize_url_as_username($url);
        }
        
        $user = create_user_record($username, $password, 'openid');
        openid_append_url($user, $url);
        
        // SREG fullname
        if (isset($sreg['fullname']) && !empty($sreg['fullname'])) {
            $name = openid_parse_full_name($sreg['fullname']);
            $user->firstname = $name['first'];
            $user->lastname = $name['last'];
        }
        
        // SREG email
        if (isset($sreg['email']) && !empty($sreg['email']) && !record_exists('user', 'email', $sreg['email'])) {
            $user->email = $sreg['email'];
        }
        
        // SREG country
        if (isset($sreg['country']) && !empty($sreg['country'])) {
            $country = $sreg['country'];
            $country_code = strtoupper($country);
            $countries = get_list_of_countries();
            
            if (strlen($country) != 2 || !isset($countries[$country_code])) {
                $countries_keys = array_keys($countries);
                $countries_vals = array_values($countries);
                $country_code = array_search($country, $countries_vals);
                
                if ($country_code > 0) {
                    $country_code = $countries_keys[$country_code];
                } else {
                    $country_code = '';
                }
            }
            
            
            if (!empty($country_code)) {
                $user->country = $country_code;
            }
        }
        
        /* We're currently not attempting to get language and timezone values
        // SREG language
        if (isset($sreg['language']) && !empty($sreg['language'])) {
        }
        
        // SREG timezone
        if (isset($sreg['timezone']) && !empty($sreg['timezone'])) {
        }
        */
        
        if (function_exists('on_openid_create_account')) {
            on_openid_create_account($resp, $user);
        }
        
        update_record('user', $user);
        $user = get_complete_user_data('id', $user->id);
        
        // Redirect the user to their profile page if not set up properly
        if (!empty($user) && user_not_fully_set_up($user)) {
            $USER = clone($user);
            $urltogo = $CFG->wwwroot.'/user/edit.php';
            redirect($urltogo);
        }
        
        return $user;
    }
}

?>
